;
; pads.s - read NES gamepads, ignoring data glitched by DPCM playback
;
; Copyright (c) 2008 Damian Yerrick
; 
; Permission is hereby granted, free of charge, to any person
; obtaining a copy of this software and associated documentation
; files (the "Software"), to deal in the Software without
; restriction, including without limitation the rights to use,
; copy, modify, merge, publish, distribute, sublicense, and/or sell
; copies of the Software, and to permit persons to whom the
; Software is furnished to do so, subject to the following
; conditions:
; 
; The above copyright notice and this permission notice shall be
; included in all copies or substantial portions of the Software.
; 
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
; OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
; WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
; OTHER DEALINGS IN THE SOFTWARE.
; 
; Visit http://www.pineight.com/ for more information.
;

JOY1      = $4016
JOY2      = $4017

; The NES has two wired game controllers, each with eight buttons.
; This subroutine reads both controllers and puts the current state
; in cur_keys[]. It also puts newly pressed keys (down this frame,
; up last frame) in new_keys[].
;
; On the NES, standard plug-in controllers show up on D0.
; But on the Famicom, the hardwired controllers show up on D0,
; and plug-in controllers show up on D1.  This function considers
; a button to be pressed if it's pressed on D0 or D1 or both.
;
; The NES CPU's I/O has a glitch.  If the DPCM playback hardware
; fetches a sample at the exact time that the program is reading
; the controller, the serial port sends an extra clock, which causes
; the program to miss one button.  But if the buttons read out the
; same way both times, this means the DPCM glitch didn't happen,
; so it's OK to update the current keys.  Worst case, we'll get
; an occasional 1-frame control lag while a key is being held.
; And unlike commercial games that loop until the buttons match
; on two successive reads, this method takes near constant time.

.export read_pads
.importzp cur_keys, new_keys
.proc read_pads

  ; Uses $0000-$0005 for local variables
  raw_keys = 0
  first_read_keys = 2
  last_keys = 4

  ; Store the current keypress state to detect key-down later
  lda cur_keys
  sta last_keys
  lda cur_keys+1
  sta last_keys+1

  ; Read the controllers twice
  jsr read_pads_once
  lda raw_keys
  sta first_read_keys
  lda raw_keys+1
  sta first_read_keys+1
  jsr read_pads_once

  ldx #1  ; For each player (1 then 0):
@fixupKeys:

    ; Compare the two reads and update only if they match
    lda raw_keys,x
    cmp first_read_keys,x
    bne @dontUpdateGlitch
      sta cur_keys,x
    @dontUpdateGlitch:

    ; Calculate newly pressed keys
    lda last_keys,x  ; A = keys that were down last frame
    eor #$FF         ; A = keys that were up last frame
    and cur_keys,x   ; A = keys down now and up last frame
    sta new_keys,x
    dex
    bpl @fixupKeys
  rts

read_pads_once:

  ; A combination data register and loop counter.
  ; When 1 is shifted left eight times, the carry will go from 0 to 1,
  ; and we're done with the loop.
  lda #$01
  sta raw_keys
  sta raw_keys+1

  ; To tell both controllers to start sending bits, send a 1
  ; then a 0 on JOY1.
  sta JOY1
  lda #$00
  sta JOY1
  loop:
    lda JOY1  ; A = port D0-D4 content of expansion port
    and #$03  ; ignore all devices except standard controllers
    cmp #$01  ; Carry = 0 if not pressed, 1 if either is pressed
    rol raw_keys
    lda JOY2
    and #$03
    cmp #$01
    rol raw_keys+1
    ; the last 'rol' instruction clocked the loop counter too
    bcc loop
  rts

.endproc
