;   t.s
;   LJ65 main code
;

;;; Copyright (C) 2005 Damian Yerrick
;
;   This program is free software; you can redistribute it and/or
;   modify it under the terms of the GNU General Public License
;   as published by the Free Software Foundation; either version 3
;   of the License, or (at your option) any later version.
;
;   This program is distributed in the hope that it will be useful,
;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;   GNU General Public License for more details.
;
;   You should have received a copy of the GNU General Public License
;   along with this program; if not, write to 
;     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;     Boston, MA  02111-1307, USA.
;
;   Visit http://www.pineight.com/ for more information.


;;; Debug settings
;   When stress testing the refresh daemon (redraw_calc plus
;   speed_draw), or debugging playfield buffer overflows,
;   turn REDRAW_ALL_THE_TIME on.
;   And use an accurate emulator.

SELECT_MAKES_NEW_PIECE = 0
REDRAW_ALL_THE_TIME = 0
DRAW_1P_STATUS = 0
PLAY_4LINE_SOUND = 1
PLAY_MOVE_SOUND = 1
DETECT_NESTICLE = 1
DRAW_OAM_AS_2P = 0
DEBUG_SPRITE_INDEX = 0
SPEED_CURVE_REPLACES_LINES = 0
LOCK_DELAY_REPLACES_LINES = 0
SKIP_TO_ATTRACT = 0


.p02


;;; Memory mapped registers

OAM       = $0200
.export OAM

PPUCTRL   = $2000
PPUMASK   = $2001
PPUSTATUS = $2002
SPRADDR   = $2003  ; always write 0 here and use DMA from OAM
PPUSCROLL = $2005
PPUADDR   = $2006
PPUDATA   = $2007

SPRDMA    = $4014
SNDCHN    = $4015


PPUCTRL_NMI      = $80
PPUCTRL_8X8      = $00
PPUCTRL_8X16     = $20
PPUCTRL_BGHIPAT  = $10
PPUCTRL_SPRHIPAT = $08
PPUCTRL_WRDOWN   = $04  ; when set, PPU address increments by 32

PPUMASK_RED      = $80  ; when set, slightly darkens other colors
PPUMASK_GREEN    = $40
PPUMASK_BLUE     = $20
PPUMASK_SPR      = $14  ; SPR: show sprites in x=0-255
PPUMASK_SPRCLIP  = $10  ; SPRCLIP: show sprites in x=8-255
PPUMASK_BG0      = $0A  ; BG0: similarly
PPUMASK_BG0CLIP  = $08
PPUMASK_MONO     = $01  ; when set, zeroes the low nibble of palette values

PPUSTATUS_VBL  = $80  ; the PPU has entered a vblank since last $2002 read
PPUSTATUS_SPR0 = $40  ; sprite 0 has overlapped BG since ???
PPUSTATUS_OVER = $20  ; More than 64 sprite pixels on a scanline since ???


;;; Some helpful macros

;;; wait4vbl
;   Waits for the PPU to signal a vertical blank.
;   Apparently local labels cause the following:
;       Internal assembler error
;       WriteExpr: Cannot write EXPR_ULABEL nodes
.macro wait4vbl
:
  bit PPUSTATUS
  bpl :-
.endmacro



;;; Constants

; game speeds are defined in terms of frames per second (FPS)
; expected speed is the approximate frame time in 3589-cycle units,
; rounded down, and the startup code checks this to make sure
; people don't cheat by playing the NTSC rom on the PAL deck

.ifdef PAL
FPS = 50
EXPECTED_SPEED = 9
.else
FPS = 60
EXPECTED_SPEED = 8
.endif
POST_CLEAR_DELAY = FPS/4
REPEAT_DELAY = FPS/5
LOCK_DELAY = FPS/2

MAX_ROWS_PER_FRAME = 11  ; half the rows that the refresh daemon can handle per vblank
MOVE_DOWN_DELAY = 2  ; at least 20 divided by MAX_ROWS_PER_FRAME
SPEEDUP_RATE_20G = 32

N_PIECES         = 7

STATE_INACTIVE          = 0
STATE_NEW_GAME          = 1
STATE_START_GAME        = 2
STATE_NEW_PIECE         = 3
STATE_FALLING_PIECE     = 4
STATE_LANDED_PIECE      = 5
STATE_CHECK_LINES       = 6
STATE_POST_CLEAR        = 7
STATE_FALL              = 8
STATE_MOVE_UP           = 9
STATE_GAMEOVER          = 10

MENU_LINE_ROTATION = 15
MENU_LINE_SPEED = 12
MENU_LINE_GARBAGE = 9
MENU_LINE_MUSIC = 6

MENU_LINE_TOP = MENU_LINE_ROTATION
MENU_LINE_BOTTOM = MENU_LINE_MUSIC
MENU_LINE_PITCH = 3
MENU_NUM_SONGS = 3

DMC_4LINES      = 0  ; position of 4 lines sound in the sample directory

; height in pixels (1/8 tiles) of the bottom of the
; bounding rectangle of the next piece 
; the piece actually occupies the top 2 rows of the
; bounding rectangle, so it's drawn 16px above this
NEXT_PIECE_HEIGHT = 149

CLEARED_ROW_TILE   = $03  ; cleared lines are flashed to this
EMPTY_ROW_TILE     = $20  ; empty lines are drawn with this
                          ; so that the scroll logic remembers which
                          ; lines it still needs to scroll down
DARK_TILE          = $10  ; game over: all blocks are flashed to this
GARBAGE_TILE       = $80  ; use bricks as garbage
GHOST_TILE         = $0F
FALLING_PIECE_TILE = $11

PF_HT = 23
PF_VIS_HT = 20

;;; Memory map
;   000-00f local variables
;   010-04f ZP static/global variables
;   090-0ff unallocated
;   180-19f palette buffer
;   1A0-1ff stack (cut that down to 32 bytes?)
;   200-2ff OAM buffer
;   300-33f music engine
;   340-3ff unallocated
;   400-4ff player 1 field
;   500-5ff player 2 field
;   600-7ff unallocated


;;; Common local variables

dividend = $0c


;;; Global variables
;   Unallocated: $6c-$ff

retrace_count = $10  ; currently, retrace_count+1 is unused but
                     ; reserved for use as retrace_count_hi
reset_time_lo = $12  ; reset_time is how long after gameover to
reset_time_hi = $13  ; go back to title screen.
rand0 = $14
rand1 = $15
rand2 = $16
rand3 = $17
rot_dir = $18
sprite_index = $19
vblanked = $1a       ; deal with NES hardware's glitchy vblank detection
cur_turn = $1b

; part of playfield that remains to be copied
dirty_bottom = $1c   ; bottom row
rows_left = $1e      ; height in rows

; game state
cur_keys = $20
new_keys = $22
das_time = $24
repress_keys = $26
das_dir = $28
ghost_y = $2A
has_floor_kicked = $2C
cur_combo = $2E
cur_state = $30
state_timer = $32
cur_flip = $34
piece_x = $36
piece_y = $38
piece_ysub = $3A
cur_piece = $3C
next_piece = $3E
score_ones = $40
score_lo = $42
score_hi = $44
lines_lo = $46
lines_hi = $48
piece_yspeed_micro = $4A
piece_yspeed = $4C
piece_yspeed_whole = $4E
rotation_system = $50
last_combo = $52
lock_delay = $54

; $56-$5C are unallocated
music_enabled = $5D
ghost_disabled = $5E
garbage_enabled = $5F
got_garbage = $60  ; Number of lines of garbage coming for this player
rand_history = $62  ; 14 bytes

; $70-$A9 are unallocated

menu_shared_dirty = $A8  ; Do shared menu items (e.g. garbage) need to be updated on this side?
well_attr_dirty = $AA
attractDataPtr = $AC  ; 2 bytes
attractCursor = $AE  ; 2 bytes
psg_sfx_state = $E0  ; $20 bytes for sound fx+music

palbuf = $180
.export palbuf
FIELD_1P = $400
FIELD_2P = $500

.exportzp cur_keys, new_keys, cur_turn, next_piece, sprite_index
.exportzp rand0, rand1, rand2, rand3, rand_history, psg_sfx_state
.exportzp cur_state, reset_time_lo, reset_time_hi, piece_yspeed, garbage_enabled
.exportzp attractDataPtr, attractCursor, dirty_bottom, state_timer
.exportzp REPEAT_DELAY, STATE_CHECK_LINES
.import read_pads, tgmstyle_init, tgmstyle_rand, random
.import init_sound, start_sound, update_sound, init_music, stop_music
.import music_playing
.import build_title_oam
.import attract_init, attract_run, attract_vblank, attract_rand, attract_pads
.import PKB_unpackblk

; I've moved rotation systems to a separate TU
; so I need to make the state of the piece available to other TUs
.import try_rotating
.export collision_check
.exportzp piece_x, piece_y, cur_flip, rot_dir, has_floor_kicked
.exportzp rotation_system, cur_piece, cur_state
.exportzp STATE_LANDED_PIECE

.segment "INESHDR"

  .byt "NES", 26
  .byt 1  ; number of 16 KB program segments
  .byt 1  ; number of 8 KB chr segments
  .byt 0  ; mapper, mirroring, etc
  .byt 0  ; extended mapper info
  .byt 0  ; number of 8 KB RAM segments
; The next two bytes are supposed to define PAL/NTSC versions
; of a given ROM, but they make Nintendulator cry.
.ifdef PAL
  .byt 0
  .byt %00000000
.else
  .byt 0
  .byt %00000000
.endif
  .byt 0
  .byt 0,0,0,0  ; DiskDude detection


.segment "CODE"

;;; Speed critical functions need to be aligned such that
;   an inner loop does not cross a page boundary.

.align 256
speed_draw:

;;; 1P
  lda rows_left
  beq @skip_1p
  ldy dirty_bottom
@loop_1p:
  lda ppubasehi,y
  beq @hidden_row_1p
  sta PPUADDR
  lda ppubaselo,y
  ora #2
  sta PPUADDR
  ldx rowOffsets,y
  .repeat 10, I
    lda FIELD_1P+I,x  ; unroll for blazing speed
    sta PPUDATA
  .endrepeat
@hidden_row_1p:
  iny
  dec rows_left
  bne @loop_1p
  sty dirty_bottom
@skip_1p:

  lda rows_left+1
  beq @skip_2p
;;; 2P
  ldy dirty_bottom+1
@loop_2p:
  lda ppubasehi,y
  beq @hidden_row_2p
  sta PPUADDR
  lda ppubaselo,y
  ora #20
  sta PPUADDR
  ldx rowOffsets,y
  .repeat 10, I
    .if DRAW_OAM_AS_2P
      lda OAM+I,x
    .else
      lda FIELD_2P+I,x  ; unroll for blazing speed
    .endif
    sta PPUDATA
  .endrepeat
@hidden_row_2p:
  iny
  dec rows_left+1
  bne @loop_2p
  sty dirty_bottom+1
@skip_2p:
  rts

nmihandler:
  inc vblanked
  rti

irqhandler:
  rti

main:

;;; Init CPU
  sei
  cld
  ldx #$40
  stx $4017
  ldx #$ff
  txs
  inx
  stx PPUCTRL
  stx PPUMASK

;;; Init machine
  lda PPUSTATUS
  wait4vbl

  ldx #0
  txa
@ramclrloop:
  sta 0,x
  sta $100,x
  sta $200,x
  sta $400,x
  sta $500,x
  sta $600,x
  sta $700,x
  inx
  bne @ramclrloop

  lda #$11
  ldy #$40
  cmp $3FD
  bne @isColdBoot
  cpy $3FE
  beq @isWarmBoot

@isColdBoot:
  sta $3FD
  sty $3FE
  lda #0
  sta $3FF
@isWarmBoot:

  ; init all variables that don't default to 0
  lda #2  ; default to song #2
  sta music_enabled

  lda #4
  sta sprite_index
  jsr fill_rest_of_sprites
  lda #$ef  ; hide sprite 0 until it's needed
  sta OAM

  wait4vbl  ; after second VBL we can write to ppu memory
  lda PPUSTATUS
  sta rand3  ; store NESTICLE-ness in bit 7 of rand3

;;; Load initial sprite table
  ldx #0
  stx SPRADDR  ;copy sprites
  lda #>OAM
  sta SPRDMA

;;; Load initial palette
  lda #$3f
  sta PPUADDR
  ldx #0
  stx PPUADDR
@palclrloop:
  lda gamepal,x
  sta palbuf,x
  inx
  cpx #32
  bne @palclrloop

  lda #$a5
  sta rand0
  jsr init_sound


;;; Load warning into nametable $2000
  .import do_intro_screens, nesticle_pkb, t_pkb, title_pkb, pause_pkb, how2_pkb
.if DETECT_NESTICLE
  lda rand3  ; if we decided it was nesticle...
  bpl @not_nesticle
  lda #<nesticle_pkb
  sta 0
  lda #>nesticle_pkb
  sta 1
  ldy #0
  jsr fbi_warning
@not_nesticle:

;jmp NTSCPALOK

; make sure you're using the right cart on the right frequency
  lda #%10000000
  sta PPUCTRL
  lda #%00000001
  sta PPUMASK
  lda #0
  sta rand0
  sta rand1
  sta vblanked
:
  lda vblanked
  beq :-
  lda #%00000000
  sta vblanked
@NTSCPALChk:
  inc rand0
  bne :+
  inc rand1
:
  lda vblanked
  beq @NTSCPALChk

  lda #EXPECTED_SPEED
  cmp rand1
  beq NTSCPALOK

; Display error message for cart used on wrong frequency system.
  jsr copy_pal
  lda #$20
  sta PPUADDR
  lda #0
  sta PPUADDR
  ldx #8
  tay
:
  sta PPUDATA
  iny
  bne :-
  dex
  bne :-

  lda #<wrong_region_message
  sta 0
  lda #>wrong_region_message
  sta 1
  lda #$21
  sta PPUADDR
  sty PPUADDR
  lda (0),y
:
  sta PPUDATA
  iny
  lda (0),y
  bne :-

  lda #0
  sta PPUSCROLL
  sta PPUSCROLL
  lda vblanked
:
  cmp vblanked
  beq :-
  lda #PPUCTRL_NMI
  sta PPUCTRL
  lda #%00011110
  sta PPUMASK
:
  jmp :-

NTSCPALOK:
.endif

.if SKIP_TO_ATTRACT
  jmp head_for_attract
.endif

  jsr do_intro_screens

title_scr:
  jsr stop_music
  tax
  sta vblanked
@hide_rainbow_band:
  lda vblanked
  beq @hide_rainbow_band
  stx PPUMASK
  jsr build_title_oam
  jsr update_sound
  ldx #3
  stx reset_time_hi
  ldx #0
  stx reset_time_lo
  lda #$20
  sta PPUADDR
  stx PPUADDR
  lda #<title_pkb
  sta 0
  lda #>title_pkb
  sta 1
  jsr PKB_unpackblk
  jsr update_sound
  lda PPUSTATUS
  lda #PPUCTRL_NMI|PPUCTRL_8X16
  sta PPUCTRL
  lda #0
  sta PPUSCROLL
  sta PPUSCROLL
  sta vblanked
@title_loop:
  lda vblanked
  beq @title_loop
  lda PPUSTATUS
  lda #0
  sta vblanked
  sta SPRADDR
  lda #>OAM
  sta SPRDMA
  lda #PPUMASK_BG0|PPUMASK_SPR
  sta PPUMASK
  jsr update_sound

  ; see if start is pressed, or if secret options are done
  jsr read_pads

  lda new_keys
  ora new_keys+1
  beq @no_keys
    lda cur_keys
    cmp #%01000010
    bne @not_B_Left
      lda #1
      sta ghost_disabled
      ldy #$0A
      lda #$80
      jsr play_dmc
    @not_B_Left:

    lda new_keys
    ora new_keys+1
    and #%00010000
    bne start_game_no_attract
  @no_keys:
  dec reset_time_lo
  bne @title_loop
  dec reset_time_hi
  bne @title_loop
  
head_for_attract:
;;; head for attract mode
  lda #STATE_INACTIVE
  sta cur_state
  lda #STATE_START_GAME
  sta cur_state+1
  lda #1
  sta attractDataPtr+1
  jmp restart_game


start_game_no_attract:
  lda #0
  sta attractDataPtr+1
  sta next_piece
  sta next_piece+1
  lda #STATE_GAMEOVER
  sta cur_state
  sta cur_state+1

restart_game:
  jsr fill_rest_of_sprites
  ldx #0
  stx PPUMASK
  stx SPRADDR  ; copy sprites
  lda #>OAM
  sta SPRDMA

  lda attractDataPtr+1
  beq @not_attract
  lda #<how2_pkb
  sta 0
  lda #>how2_pkb
  sta 1
  lda #$20
  sta PPUADDR
  stx PPUADDR
  jsr PKB_unpackblk
  jmp done_loading_pf_bg

@not_attract:
;;; bring in gameplay screen
  lda #<t_pkb
  sta 0
  lda #>t_pkb
  sta 1
  lda #$20
  sta PPUADDR
  stx PPUADDR
  jsr PKB_unpackblk
  ldx #1
  jsr loadWellAttributeNow
  ldx #0
  jsr loadWellAttributeNow
;;; fill it vertically
  lda ppubasehi+19
  sta PPUADDR
  lda ppubaselo+19
  sta PPUADDR
  ldy #PF_VIS_HT
@mk_pf_borders:
  ldx #0
:
  lda pf_side_borders,x
  sta PPUDATA
  inx
  cpx #32
  bcc :-
  dey
  bne @mk_pf_borders

  ldx #0
  lda #<pause_pkb
  sta 0
  lda #>pause_pkb
  sta 1
  lda #$2c
  sta PPUADDR
  stx PPUADDR
  jsr PKB_unpackblk


done_loading_pf_bg:
  ldx #0
:
  lda gamepal,x
  sta palbuf,x
  inx
  cpx #32
  bne :-

  lda #0
  sta vblanked
  sta garbage_enabled
  sta rotation_system
  sta rotation_system+1
  sta piece_y
  sta piece_y+1
  lda #1
  sta state_timer
  sta state_timer+1

  ; turn the screen on for the first time
:
  lda vblanked
  beq :-
  lda PPUSTATUS
  lda #PPUCTRL_NMI
  sta PPUCTRL
  lda #0
  sta PPUSCROLL
  sta PPUSCROLL
  lda #%00011110
  sta PPUMASK


;
; THE MAIN GAME LOOP
;
game_loop:
  inc retrace_count
  jsr read_pads
  lda attractDataPtr+1
  beq :+
    jsr attract_pads
    jmp @no_pause
  :

  ; handle pause

  lda cur_state
  cmp #STATE_NEW_PIECE
  bcs @someone_can_pause
  lda cur_state+1
  cmp #STATE_NEW_PIECE
  bcc @no_pause
@someone_can_pause:
  lda new_keys
  ora new_keys+1
  and #%00010000
  beq @no_pause

    ;display pause screen
    lda music_playing
    pha
    lda #%00001010
    sta PPUMASK
    lda #3
    ldy #0
    sty music_playing
    jsr startwait
    lda #0
    sta PPUCTRL
    sta PPUSCROLL
    sta PPUSCROLL
    lda #%00011110
    sta PPUMASK
    pla
    sta music_playing
    ldx cur_turn
  @no_pause:



  ldx #0
  jsr game_cycle
  ldx #1
  jsr game_cycle
  jsr update_sound

  ; if we've run out of next pieces, leave attract mode
  lda next_piece+1
  bpl @no_leave_attract
    jmp title_scr
  @no_leave_attract:

  ldx #0
  jsr draw_next
  ldx #1
  jsr draw_next

  ldx #0
  jsr draw_piece
  ldx #1
  jsr draw_piece

  ldx #0
  jsr draw_ghost
  ldx #1
  jsr draw_ghost

  lda attractDataPtr+1
  beq @notAttract
    jsr attract_run
  @notAttract:

  jsr fill_rest_of_sprites
  jsr redraw_calc

  ; wait for sprite 0 hit flag to become false at end of vblank,
  ; in case processing completed entirely within vblank
:
  bit PPUSTATUS
  bvs :-

; wait for sprite 0 hit flag to become true before start of vblank
; with a failsafe in case sprite 0 is missed
:
  bit PPUSTATUS  ; wait for sprite 0 hit
  bmi @vblank_instead
  bvc :-

;;; BEGIN VRAM UPDATE CODE
;   (optimize the crap out of this)

  ; On 2008-12-22, blargg found a glitch in the PPU.
  ; If you turn off rendering on a row containing a sprite,
  ; or at any horizontal time other than x=64 to 255,
  ; OAM may drop sprites during the next frame.
  ; So put the solid part of sprite 0 on its last scanline,
  ; and wait for roughly the center of the screen before
  ; turning off rendering.
  ldy #10
:
  dey
  bne :-
  sty vblanked
  sty PPUMASK
  ldy #PPUCTRL_NMI
  sty PPUCTRL

  ; Copy attract mode help text to the screen if needed.
  ; Do it BEFORE copying the playfield because it needs to know
  ; whether redraw_calc scheduled a playfield update.
  lda attractDataPtr+1
  beq @notAttractVblank
    jsr attract_vblank
  @notAttractVblank:
  jsr loadWellAttribute

  jsr speed_draw

  ; When rendering is turned off, and PPUADDR is pointed at the palette,
  ; the PPU shows the color that PPUADDR points to instead of the
  ; background color.  So wait for vblank before updating the palette.
:
  lda vblanked
  beq :-
@vblank_instead:

  lda PPUSTATUS
  lda #0
  sta PPUMASK
  jsr copy_pal
  lda #0
  sta PPUSCROLL
  sta PPUSCROLL
  sta SPRADDR  ; copy sprites
  ldx #>OAM
  stx SPRDMA
  lda #PPUCTRL_NMI
  sta PPUCTRL
  lda #%00011110
  sta PPUMASK
;;; END IN-GAME VBLANK CODE

  lda cur_state  ; return to title screen if nothing has happened
  bne @bunny
  lda cur_state+1
  beq @countdown_to_title
@bunny:  ; still going?
  lda #$fc
  sta reset_time_hi
  lda #0
  sta reset_time_lo
@jgl:
  jmp game_loop

@countdown_to_title:
  inc reset_time_lo
  bne @jgl
  inc reset_time_hi
  bne @jgl
  jmp title_scr

game_cycle:
  stx cur_turn
  lda new_keys,x
  eor #$FF
  and repress_keys,x
  sta repress_keys,x

  ; if in attract mode, and start has been pressed,
  ; kill the next piece in order to leave attract mode
  lda attractDataPtr+1
  beq @not_attract_exit
    lda new_keys,x
    and #%00010000
    beq @not_attract_exit
    lda #$FF
    sta next_piece+1
    lda #0
    sta cur_state+1
    rts
  @not_attract_exit:

  ; if left or right was just pressed,
  ; set that as the new DAS direction
  lda new_keys,x
  and #%00000011
  beq @no_set_dir
    sta das_dir,x
    lda #0
    sta das_time,x
  @no_set_dir:

  ; if the DAS direction is still held down,
  ; charge the DAS
  lda cur_keys,x
  and das_dir,x
  beq @das_skip
    inc das_time,x
    lda das_time,x

    ; if the DAS is sufficiently charged,
    ; simulate a keypress in the DAS direction
    cmp #REPEAT_DELAY
    bcc @das_skip
    lda cur_keys,x
    and das_dir,x
    ora new_keys,x
    sta new_keys,x
    dec das_time,x
  @das_skip:

  ; jump table for player state

  lda cur_state,x
  asl a
  tax
  lda state_jtable+1,x
  pha
  lda state_jtable,x
  pha
straight_rts:
  rts

.align 2
state_jtable:
  .addr wait_for_join-1
  .addr do_new_game_menu-1
  .addr make_new_game-1
  .addr make_new_piece-1
  .addr falling_piece-1
  .addr falling_piece-1
  .addr find_full_lines-1
  .addr post_clear-1
  .addr move_down-1
  .addr move_up-1
  .addr game_over_anim-1

wait_for_join:
  lda attractDataPtr+1
  bne straight_rts
  ldx cur_turn
  lda cur_keys,x
  and #%11000000
  cmp #%11000000  ; check that A and B are both down
  bne straight_rts
  and new_keys,x  ; also check that one of them was just pressed
  beq straight_rts

init_new_game_menu:
  ; if the tempo is zero, then music isn't already playing,
  ; so turn on the music
  lda music_playing
  bne @music_already_playing
  lda music_enabled
  beq @music_already_playing
  sec
  sbc #1
  jsr init_music
@music_already_playing:

  lda cur_turn
  asl a
  asl a
  tax
  lda palbuf+7,x
  sta palbuf+19,x
  ldx cur_turn
  txa
  ora #>FIELD_1P
  sta 1
  ldy #0
  sty 0
  ldy #199
:
  lda menu_text,y
  sta (0),y
  dey
  bne :-
  lda menu_text
  sta (0),y
  ldy #10*PF_HT
  sty 0
  ldy #19
:
  lda menu_text+200,y
  sta (0),y
  dey
  bpl :-

  lda #0
  sta dirty_bottom,x
  lda #STATE_NEW_GAME
  sta cur_state,x
  lda #MENU_LINE_TOP
  sta piece_y,x
  lda #0
  sta piece_yspeed,x
  jsr update_menu_rotation
  jsr update_menu_speed
  jmp update_menu_shared

do_new_game_menu:
  ldx cur_turn
  txa
  ora #>FIELD_1P
  sta 1
  ldy #0
  sty 0
  lda menu_shared_dirty,x
  beq :+
  jsr update_menu_shared
:
  lda new_keys,x
  and #%10000000
  beq @not_A
  jmp make_new_game
@not_A:

  lda new_keys,x
  and #%00001000
  beq @not_up
  lda piece_y,x
  cmp #MENU_LINE_TOP
  bcs @not_up
  adc #MENU_LINE_PITCH
  sta piece_y,x
@not_up:

  lda new_keys,x
  and #%00000100
  beq @not_down
  lda piece_y,x
  cmp #MENU_LINE_BOTTOM + MENU_LINE_PITCH
  bcc @not_down
  sbc #MENU_LINE_PITCH
  sta piece_y,x
@not_down:

  lda new_keys,x
  and #%00000010
  beq @not_left
  lda piece_y,x
  sta dirty_bottom,x
  cmp #MENU_LINE_ROTATION
  beq toggle_rotation
  cmp #MENU_LINE_MUSIC
  bne @not_prev_song
  lda #MENU_NUM_SONGS - 1
  jmp change_music
@not_prev_song:
  cmp #MENU_LINE_GARBAGE
  beq toggle_garbage
  cmp #MENU_LINE_SPEED
  bne @not_left
  dec piece_yspeed,x
  jmp update_menu_speed
@not_left:

  lda new_keys,x
  and #%00000001
  beq @not_right
  lda piece_y,x
  sta dirty_bottom,x
  cmp #MENU_LINE_ROTATION
  beq toggle_rotation
  cmp #MENU_LINE_MUSIC
  bne @not_next_song
  lda #1
  jmp change_music
@not_next_song:
  cmp #MENU_LINE_GARBAGE
  beq toggle_garbage
  cmp #MENU_LINE_SPEED
  bne @not_right
  inc piece_yspeed,x
  jmp update_menu_speed
@not_right:

  rts

toggle_rotation:
  lda rotation_system,x
  eor #1
  sta rotation_system,x
update_menu_rotation:
  lda rotation_system,x
  asl a
  asl a
  asl a
  ora #7
  tax
  lda #MENU_LINE_ROTATION*10+1
  sta 0
  ldy #7
@write_name:
  lda rotation_names,x
  sta (0),y
  dex
  dey
  bpl @write_name
  ldx cur_turn
  rts

change_music:
  clc
  adc music_enabled
  cmp #MENU_NUM_SONGS
  bcc @not_music_wrap
  sbc #MENU_NUM_SONGS
  clc
@not_music_wrap:
  ; at this point, the carry is clear
  sta music_enabled
  sta music_playing
  beq :+
    adc #$FF  ; subtract 1
    jsr init_music
  :
  lda #1
  sta menu_shared_dirty
  sta menu_shared_dirty+1
  jmp update_menu_shared

toggle_garbage:
  lda garbage_enabled
  eor #1
  sta garbage_enabled
  lda #1
  sta menu_shared_dirty
  sta menu_shared_dirty+1
update_menu_shared:
  lda garbage_enabled
  asl a
  asl a
  ora #3
  tax
  lda #MENU_LINE_GARBAGE*10+5
  sta 0
  ldy #3
@write_garbage_name:
  lda on_off_text,x
  sta (0),y
  dex
  dey
  bpl @write_garbage_name

  lda music_enabled
  asl a
  asl a
  asl a
  ora #7
  tax
  lda #MENU_LINE_MUSIC*10+1
  sta 0
  ldy #7
@write_song_name:
  lda song_names,x
  sta (0),y
  dex
  dey
  bpl @write_song_name

  ldx cur_turn
  lda #0
  sta menu_shared_dirty,x
  lda #MENU_LINE_MUSIC
  cmp dirty_bottom,x
  bcs :+
  sta dirty_bottom,x
:
  lda #0
  sta menu_shared_dirty,x
  sta 0
  rts

update_menu_speed:
  lda piece_yspeed,x
  and #%00000011
  sta piece_yspeed,x
  asl a
  asl a
  ora #3
  tax
  lda #MENU_LINE_SPEED*10+5
  sta 0
  ldy #3
@write_name:
  lda speed_names,x
  sta (0),y
  dex
  dey
  bpl @write_name
  ldx cur_turn
  rts

make_new_game:

  lda attractDataPtr+1
  beq @not_attract
    jsr attract_init
    jmp @done_attract
  @not_attract:
    jsr tgmstyle_init
  @done_attract:

  ldx cur_turn
  txa           ; clear out playfield
  ora #>FIELD_1P
  sta 1
  ldy #0
  sty 0
  tya
@clear_field:
  sta (0),y
  iny
  sta (0),y
  iny
  cpy #PF_HT * 10
  bcc @clear_field
  sty 0

  ; init the score only if there isn't an attract sequence
  ldy attractDataPtr+1
  bne @no_copy_score
@copy_score:
  lda board_top,y
  sta (0),y
  iny
  lda board_top,y
  sta (0),y
  iny
  cpy #20
  bcc @copy_score
@no_copy_score:

  lda #0
  sta lines_lo,x
  sta lines_hi,x
  sta score_ones,x
  sta score_lo,x
  sta cur_combo,x
  sta score_hi,x
  sta got_garbage,x
  
  lda piece_yspeed,x  ; set initial speed
  tay
  lda initial_speeds,y
  sta piece_yspeed,x
  lda initial_speeds_whole,y
  sta piece_yspeed_whole,x
  lda #0
  sta dirty_bottom,x
  sta piece_yspeed_micro,x

  lda attractDataPtr+1
  beq @not_attract_rand
    jsr attract_rand
    ldx cur_turn
    jmp @done_rand
  @not_attract_rand:
    jsr tgmstyle_rand
    sta next_piece,x

    ; only set the well color scheme in non-attract mode
    ; (the attract mode nametable data takes care of it
    ; in attract mode)
    lda #1
    sta well_attr_dirty,x
  @done_rand:

  lda #STATE_NEW_PIECE
  sta cur_state,x
  
  ; wait a half second before starting the first piece
  lda #LOCK_DELAY
  sta state_timer,x
  jmp set_piece_colors

.import darkcolors, fallencolors, segaColorScramble

make_new_piece:

  ; handle ARE (entry delay)
  ldx cur_turn
  lda state_timer,x
  beq @are_expired
  dec state_timer,x
  beq @are_expired
  rts
@are_expired:
  
  jsr update_score
  ldx cur_turn

  ; update piece speed
  ; but not if it's already 20G
  lda piece_yspeed_whole,x
  cmp #20
  bcs @speedup_20g

  sta 1
  lda #LOCK_DELAY  ; FIXME: slowdown for debug purposes.
  sta lock_delay,x
  lda piece_yspeed,x
  asl a
  rol 1
  ; 1:A holds speed*2
  clc
  adc piece_yspeed,x
  tay
  lda 1
  adc piece_yspeed_whole,x
  sta 1
  ; 1:Y holds speed*3
  ; so now add speed*3/256 to speed

  tya
  adc piece_yspeed_micro,x
  sta piece_yspeed_micro,x
  lda piece_yspeed,x
  adc 1
  sta piece_yspeed,x
  bcc @no_speedup
  inc piece_yspeed_whole,x
  jmp @no_speedup
@speedup_20g:
  lda #SPEEDUP_RATE_20G
  clc
  adc piece_yspeed,x
  sta piece_yspeed,x
  bcc :+
  inc piece_yspeed_whole,x
:
  lda #1
  sta 2         ; quotient (use carry as loop terminator)
  lda piece_yspeed_whole,x
  sta 1         ; divisor (denominator)
  lda #0
  lda #<(FPS*10)
  sta 0         ; dividend (numerator)
  lda #>(FPS*10)
@divloop:
  asl 0
  rol a
  bcs @is_greater
  cmp 1
  bcc @not_greater
@is_greater:
  sbc 1
  sec
@not_greater:
  rol 2
  bcc @divloop
  lda 2
  sta lock_delay,x

@no_speedup:
  lda next_piece,x
  sta cur_piece,x

  lda attractDataPtr+1
  beq @not_attract_rand
    jsr attract_rand
    ldx cur_turn
    jmp @done_rand
  @not_attract_rand:
    jsr tgmstyle_rand
    sta next_piece,x
  @done_rand:

  lda #0
  sta cur_flip,x
  sta has_floor_kicked,x

  ; Initial rotation system ;-)
  lda #%00111111
  and cur_keys,x
  sta cur_keys,x

  lda #PF_VIS_HT - 2
  sta piece_y,x

  ; in Sega, turn JLSTZ upside down and move up one row
  lda rotation_system,x
  beq @spawn_not_sega
  lda cur_piece,x
  beq @spawn_not_sega
  cmp #3
  beq @spawn_not_sega
  inc piece_y,x
  lda #2
  sta cur_flip,x
@spawn_not_sega:

  lda #3
  sta piece_ysub,x
  sta piece_x,x
  jsr collision_check
  beq @not_spawn_over_blocks
    jsr set_piece_colors
    jmp set_game_over
@not_spawn_over_blocks:
  lda #STATE_FALLING_PIECE
  sta cur_state,x
  lda #0
  sta state_timer,x
  ; disable down until it is repressed
  lda repress_keys,x
  ora #%00000100
  sta repress_keys,x

set_piece_colors:
  ; get the color of the current piece
  lda cur_piece,x
  tay
  lda rotation_system,x
  beq @cur_not_sega
  lda segaColorScramble,y
  tay
@cur_not_sega:
  txa
  asl a
  asl a
  tax   ; palette offset = 4 * x
  lda darkcolors,y
  sta palbuf+17,x
  adc #$10
  sta palbuf+18,x
  adc #$10
  sta palbuf+19,x

  ; get the color of the next piece
  ldx cur_turn
  lda next_piece,x
  tay
  lda rotation_system,x
  beq @next_not_sega
  lda segaColorScramble,y
  tay
@next_not_sega:
  txa
  asl a
  asl a
  tax
  lda darkcolors,y
  sta palbuf+25,x
  adc #$10
  sta palbuf+26,x
  adc #$10
  sta palbuf+27,x
  jmp update_ghost


falling_piece:
  ldx cur_turn

.if SELECT_MAKES_NEW_PIECE
  lda new_keys,x
  and #%00100000
  beq @no_mknewpiece
  jmp make_new_piece
@no_mknewpiece:
.endif

  ; use rot_dir to see if the ghost piece needs to be updated
  lda #0
  sta rot_dir

  ;do control and collision here
  lda new_keys,x
  and #%01000000  ; B
  beq @no_lflip
  lda #3
  jsr try_rotating
  ldx cur_turn
@no_lflip:
  lda new_keys,x
  and #%10000000  ; A
  beq @no_rflip
  lda #1
  jsr try_rotating
  ldx cur_turn
@no_rflip:
  lda new_keys,x
  and #%00000010  ; Left
  beq @no_left
  lda #-1
  jsr try_shifting
  ldx cur_turn
@no_left:
  lda new_keys,x
  and #%00000001  ; Right
  beq @no_right
  lda #1
  jsr try_shifting
  ldx cur_turn
@no_right:

  ; If and only if the player tried to shift or rotate the piece,
  ; update the landed state and ghost piece position.
  ; FIXME: For speed, actually apply this only if the shift or
  ; rotate was successful.
  lda rot_dir
  beq @no_update_ghost
  jsr check_still_landed
  jsr update_ghost
@no_update_ghost:

  ; Now that we've updated the ghost piece position,
  ; move the piece down by whole rows at a time if possible.
  lda piece_y,x
  sec
  sbc piece_yspeed_whole,x
  cmp ghost_y,x
  bpl @not_landing_after_whole_rows
  lda #0
  sta piece_ysub,x
  lda ghost_y,x
@not_landing_after_whole_rows:
  sta piece_y,x

  lda repress_keys,x
  eor #$FF
  and cur_keys,x
  and #%00000100  ; Down
  beq @no_drop
  lda cur_state,x
  cmp #STATE_LANDED_PIECE
  bne @soft_drop

  ; here we know that the piece has landed and the player
  ; is holding Down.  In CENTER, the player has to re-press
  ; Down to lock the piece, but in BOTTOM, soft drop locks
  ; on contact.
  lda rotation_system,x
  bne @locking_soft_drop
  lda new_keys,x
  and #%00000100  ; Down
  beq @no_drop
@locking_soft_drop:
  jmp @lock_piece

@soft_drop:

  lda #0
  sta piece_ysub,x
  inc score_ones,x
  lda score_ones,x
  cmp #100
  bcc :+
  sbc #100
  sta score_ones,x
  inc score_lo,x
  bne :+
  inc score_hi,x
:
  jsr update_score
@no_drop:

  ldx cur_turn
  lda new_keys,x
  and #%00001000
  beq @no_harddrop

  lda cur_state,x
  cmp #STATE_LANDED_PIECE
  beq @lock_piece

  ; so we have a hard drop.
  ; now figure score as distance+1
  lda piece_y,x
  sec
  sbc ghost_y,x
  sec  ; this carry adds the plus 1
  adc score_ones,x
  cmp #100
  bcc @no_ones_ovf
  sbc #100
  inc score_lo,x
  bne @no_ones_ovf
  inc score_hi,x
@no_ones_ovf:
  sta score_ones,x

  ; In BOTTOM (Ti-ARS), hard drop has a lock delay.
  ; In CENTER (SRS-style), lock delay after hard drop is one frame.
  lda #1
  ldy rotation_system,x
  beq @using_sonic_lock
  lda lock_delay,x
@using_sonic_lock:
  sta state_timer,x
  lda #STATE_LANDED_PIECE
  sta cur_state,x
  lda #0
  sta piece_ysub,x
  lda ghost_y,x
  sta piece_y,x
  lda #2
  jsr start_sound
  ldx cur_turn
@no_harddrop:

  ; If the piece is ready to fall into the next row,
  ; move it down.
  lda piece_ysub,x
  sec
  sbc piece_yspeed,x
  sta piece_ysub,x
  bcc @drop_by_1

  ; Otherwise, test if the piece has just landed through
  ; falling whole numbers of blocks.
  lda ghost_y,x
  cmp piece_y,x
  bne @no_lock
  lda piece_ysub,x
  bne @no_lock
  beq @set_landed

@drop_by_1:
  dec piece_y,x

  jsr collision_check  ; FIXME: we can apply ghost_y here
  beq @no_lock
  ldx cur_turn
  inc piece_y,x
  lda #0
  sta piece_ysub,x

@set_landed:
  lda #STATE_LANDED_PIECE
  cmp cur_state,x
  beq @lock_countdown
  sta cur_state,x
  lda lock_delay,x
  sta state_timer,x
  lda #2
  jsr start_sound
  ldx cur_turn
  rts

@lock_countdown:
  dec state_timer,x
  beq @lock_piece
  rts
@lock_piece:
  lda #3
  jsr start_sound
  jsr commit_piece

@no_lock:
  rts


;;
; Finds where the piece would land if hard-dropped.
; Because it has to check all blocks of the piece against all
; blocks below it in the field, this can be a VERY expensive
; function in CPU time.
update_ghost:
  ldx cur_turn
  lda piece_y,x
  sta 7  ; 7: y position before hard drop
@harddrop_loop:
  dec piece_y,x
  jsr collision_check
  beq @harddrop_loop   ; while there's no collision, keep looking
  inc piece_y,x
  lda piece_y,x
  sta ghost_y,x
  lda 7
  sta piece_y,x
  rts

.import xshapes, yshapes

;
; temp variables used:
;   0 piece y
;   1 piece_x
;   3 number of blocks left
;   4-5 pointer to field
; trashes 2, 3
; out:
;   !Z = was there a collision found
;   6 = >0 if all collisions were in the central column, or
;       0 if not (ARS criterion for JLSTZ kickability)
;   
collision_check:
  ldx cur_turn
  txa
  ora #>FIELD_1P
  sta 5
  lda #0
  sta 4
  sta 6

  ; calculate piece y
  lda piece_y,x
  sta 0

  lda piece_x,x
  sta 1

  lda cur_piece,x
  asl a
  asl a
  ora cur_flip,x
  asl a
  asl a
  tax

  lda #4
  sta 3

@blockloop:
  lda xshapes,x
  clc
  adc 1
  cmp #10       ; x<0 or x>=10: collision
  bcs @collision_found
  sta 2

  lda yshapes,x
  adc 0         ; y<0: collision
  bmi @collision_found
  cmp #PF_HT
  bcs @above         ; y>=20: no collision
  tay
  lda rowOffsets,y
  adc 2
  tay
  lda (4),y     ;load the block
  bmi @collision_found  ; bit 7: solid
@above:
  inx
  dec 3
  bne @blockloop
;no collision found; restore registers
  ldx cur_turn
  lda 6
  rts

@collision_found:
  ; In BOTTOM (Ti-ARS) rotation, a JLSTZ piece can be kicked
  ; only if a block outside the piece's central column overlaps
  ; the wall or the stack.  So now we count central column
  ; collisions in 6 and keep checking the rest of the piece for
  ; collisions.  If 6 is nonzero at the end of the piece, there was
  ; a collision.  But any collisions not in the central column will
  ; stop the check, zero out 6 (indicating kickability), and return
  ; a collision immediately.
  lda xshapes,x
  cmp #1
  bne @not_central_column
  inc 6
  bne @above
@not_central_column:
  ldx cur_turn
  lda #0
  sta 6
  lda #1
  rts


; try_shifting
; input:
; cur_turn = player number
; piece_x,x, piece_y,x = current position of piece
try_shifting:
  ldx cur_turn
  sta rot_dir
  clc
  adc piece_x,x
  sta piece_x,x
  jsr collision_check
  beq @finish_shift
  lda piece_x,x
  sec
  sbc rot_dir
  sta piece_x,x
  rts

@finish_shift:
.if PLAY_MOVE_SOUND
  lda #1
  jsr start_sound
.endif
  rts

; check_still_landed
; If a piece has moved, call this to update its landed/falling state.
check_still_landed:
  ldx cur_turn
  lda #STATE_LANDED_PIECE
  cmp cur_state,x
  bne @wasnt_landed

  ;State is landed.  Decide whether to stay in landed
  ;or to leave landed state.
  dec piece_y,x
  jsr collision_check
  beq @not_resting

  ; Still resting on something.
  inc piece_y,x
  rts

@not_resting:
  ; Move piece to top of this row.
  lda #$FF
  sta piece_ysub,x
  lda #STATE_FALLING_PIECE
  sta cur_state,x
@wasnt_landed:
  rts




;;
; Writes the piece at (piece_x, piece_y, cur_flip)
; to the screen.
; Clobbers 0-5.
; out: 6 = number of blocks locked below ceiling
commit_piece:
  lda cur_turn
  asl a
  asl a
  tax
  lda #$30
  sta palbuf+17,x
  sta palbuf+18,x
  sta palbuf+19,x
  ldx cur_turn
  lda #STATE_CHECK_LINES
  sta cur_state,x
  txa
  ora #>FIELD_1P
  sta 5
  lda #0
  sta 4
  sta 6

  lda piece_y,x         ; redraw starting at the bottom of the piece
  sta 0
  bpl @not_below        ; make sure not redrawing below playfield
  lda #0
@not_below:
  sta dirty_bottom,x

  lda piece_x,x
  sta 1

  lda cur_piece,x
  tay
  lda rotation_system,x
  beq @commit_not_sega
  lda segaColorScramble,y
  tay
@commit_not_sega:
  lda fallencolors,y
  sta 2

  lda cur_piece,x
  asl a
  asl a
  ora cur_flip,x
  asl a
  asl a
  tax

  lda #4
  sta 3

@commit_loop:
  lda yshapes,x
  clc
  adc 0         ; (piece_y + blk_y)
  cmp #PF_HT
  bpl set_game_over ;y>=top of pf: Game Over!
  cmp #PF_VIS_HT
  bpl @not_below_ceiling
  inc 6
@not_below_ceiling:
  tay
  clc            ; without this, all blocks in vanish zone are
                 ; shifted right by 1
  lda rowOffsets,y ;y * 10
  adc xshapes,x  ;  + x
  adc 1          ;  + piece_x
  tay
  lda 2
  sta (4),y     ;fill in the block
  inx
  dec 3
  bne @commit_loop

  ; if no pieces were locked below the ceiling, game over
  lda 6
  beq set_game_over
  rts

set_game_over:
  ldx cur_turn
  lda #PF_VIS_HT - 1
  sta piece_y,x
  lda #STATE_GAMEOVER
  sta cur_state,x
  lda #1
  sta state_timer,x
  sta well_attr_dirty,x

  lda #7
  jsr start_sound
  lda #8
  jsr start_sound

;;; darken blocks
  ldx cur_turn
  ldy #0
  sty dirty_bottom,x
  sty 4
@darken_loop:
  lda (4),y
  beq @is_empty
  lda #DARK_TILE
  sta (4),y
@is_empty:
  iny
  cpy #PF_VIS_HT*10
  bne @darken_loop
  rts


game_over_anim:

  ; kill the music only if other player isn't playing
  lda cur_turn
  eor #1
  tax
  lda cur_state,x
  beq @killmusic
  cmp #STATE_GAMEOVER
  bne @notkillmusic
@killmusic:
  jsr stop_music
@notkillmusic:

  ldx cur_turn

  ; Don't advance until the well color scheme is updated
  lda well_attr_dirty,x
  bne @skip

  dec state_timer,x
  bne @skip

@ready:
  txa
  ora #>FIELD_1P
  sta 5
  lda #0
  sta 4
  sta dirty_bottom,x
  lda piece_y,x
  tay
  lda rowOffsets,y
  tax
  ldy #0
@loop:
  lda gameover_text,x
  sta (4),y
  inx
  iny
  lda gameover_text,x
  sta (4),y
  inx
  iny
  cpx #PF_VIS_HT * 10
  bcc @loop
  ldx cur_turn
  dec piece_y,x
  bpl @still_going
  lda #STATE_INACTIVE
  sta cur_state,x
  rts
@still_going:
  lda #MOVE_DOWN_DELAY
  sta state_timer,x
@skip:
  rts


draw_ghost:
  stx cur_turn
  lda ghost_disabled
  bne @no_draw_ghost
  lda cur_state,x
  cmp #STATE_FALLING_PIECE
  beq @is_falling
@no_draw_ghost:
  rts
@is_falling:
  stx 2  ; draw ghost piece with falling piece palette
  lda #GHOST_TILE
  sta 5  ; draw ghost with the dark tile

  ; calculate piece y
  lda ghost_y,x
  asl a
  asl a
  asl a
  sta 0         ;0: bottom of piece, in pixels
  jmp draw_piece_in_field

draw_next:
  lda #FALLING_PIECE_TILE
  sta 5
  stx cur_turn
  lda cur_state,x
  cmp #STATE_NEW_PIECE
  bcs @not_inactive
  rts
@not_inactive:
  txa
  ora #2
  sta 2

  lda #NEXT_PIECE_HEIGHT
  sta 0

  ; in Sega, turn JLSTZ upside down and move up one row
  lda #0
  sta 3  ; initial orientation
  lda rotation_system,x
  beq @next_not_sega
  lda next_piece,x
  beq @next_not_sega
  cmp #3
  beq @next_not_sega
  lda 0
  clc
  adc #8
  sta 0
  lda #2
  sta 3
@next_not_sega:

  lda #PF_HT * 10
  sta 4
  lda #3
  clc
  adc board_x,x
  asl a
  asl a
  asl a
  sta 1
  lda next_piece,x
  asl a
  asl a
  ora 3
  jmp draw_a_piece

draw_piece:
  stx cur_turn

  lda cur_state,x       ;if the piece isn't active don't draw it
  cmp #STATE_FALLING_PIECE
  beq @is_falling
  cmp #STATE_LANDED_PIECE
  beq @is_falling
  cmp #STATE_CHECK_LINES  ; for lock flash
  beq @is_falling
  cmp #STATE_NEW_GAME
  beq @is_new_game_menu
  rts

;;; In the new game menu, the player's "piece" is an arrow at
;   the left side that bounces back and forth on the line.
@is_new_game_menu:
  lda piece_y,x
  asl a
  asl a
  asl a
  eor #$ff
  clc
  adc #216
  sta 0
  lda board_x,x
  asl a
  asl a
  asl a
  sta 1
  lda #%00010000  ; compute bouncing arrow
  and retrace_count
  beq :+
  lda #$0f
:
  eor retrace_count
  and #%00001110
  lsr a
  adc 1
  sta 1

  ldx sprite_index
  lda 0
  sta $200,x
  inx
  lda #31  ; greater-than tile
  sta $200,x
  inx
  lda cur_turn
  sta $200,x
  inx
  lda 1
  sta $200,x
  inx
  stx sprite_index
  ldx cur_turn
  rts

@is_falling:
  lda #FALLING_PIECE_TILE
  sta 5
  stx 2                 ; 2: palette to draw the piece in

  ; calculate piece y
  lda piece_y,x
  asl a
  asl a
  asl a
  sta 0
  lda piece_ysub,x
  lsr a
  lsr a
  lsr a
  lsr a
  lsr a
  ora 0
  sta 0         ;0: bottom of piece, in pixels

draw_piece_in_field:

  lda #PF_VIS_HT * 8    ;don't draw blocks above the top row
  sta 4

  ; calculate piece x
  lda piece_x,x
  clc
  adc board_x,x
  asl a
  asl a
  asl a
  sta 1         ;1: left side of piece, in pixels

  ; calculate base addr of piece data
  lda cur_piece,x
  asl a
  asl a
  ora cur_flip,x

;;;
; Tail-called from draw_piece and draw_next.
; @param A shape of current piece * 4 + rotation
; @param 0 bottom of piece, in pixels
; @param 1 X coord of left side of current piece in pixels
; @param 2 sprite palette id used to draw the piece
;          (player number, plus 0 if falling or 2 if next)
; @param 4 highest Y to draw a piece
draw_a_piece:
  asl a
  asl a           ;y = piece data offset
  tay             ;  = 16 * piece + 4 * rotation

  lda #4          ;3 = remaining number of blocks in piece
  sta 3

  ldx sprite_index
@blockloop:
  lda yshapes,y
  asl a
  asl a
  asl a
  adc 0
  cmp 4
  bcs @above_field
  eor #$ff
  clc
  adc #216
  sta OAM,x
  inx
  lda 5        ; tile in which to draw the piece
  sta OAM,x
  inx
  lda 2  ; vhp000cc = 001000np where p = turn and n = next
  ora #%00100000  ; draw falling blocks behind frame
  sta OAM,x
  inx
  lda xshapes,y
  asl a
  asl a
  asl a
  adc 1
  sta OAM,x
  inx
@above_field:
  iny
  dec 3
  bne @blockloop
  
; debugging sh**
.if DRAW_1P_STATUS
  lda #119
  sta $200,x
  inx
  lda piece_y
  ora #$30
  sta $200,x
  inx
  lda #0
  sta $200,x
  inx
  lda #120
  sta $200,x
  inx
  lda #127
  sta $200,x
  inx
  lda cur_state
  ora #$30
  sta $200,x
  inx
  lda #0
  sta $200,x
  inx
  lda #120
  sta $200,x
  inx
.endif
  stx sprite_index
  rts


fill_rest_of_sprites:
  lda sprite_index
  and #%11111100
  tax
  .if DEBUG_SPRITE_INDEX
  lda #64
  sta OAM,x
  sta OAM+1,x
  sta OAM+3,x
  lda #0
  sta OAM+2,x
  inx
  inx
  inx
  inx
  .endif
  lda #$f0
@loop:
  sta OAM,x
  inx
  inx
  inx
  inx
  bne @loop

  ; setup up sprite 0 at (248, 223)
  ; (see comment about blargg's OAM glitch above)
  lda #222
  sta OAM
  lda #'_'
  sta OAM+1
  lda #%00100011  ; behind
  sta OAM+2
  lda #248
  sta OAM+3
  ldx #4
  stx sprite_index
  rts


;
; find_full_lines
; For player x, look for filled lines.  Leave the number of such lines
; in 2.
;

.import play_dmc

find_full_lines:
  jsr erasePFBGText
  ldx cur_turn
  txa
  ora #>FIELD_1P
  sta 5
  lda #0
  sta 4

  lda piece_y,x       ; current line
  bpl @not_below      ; clip against bottom
  lda #0
@not_below:
  cmp #PF_VIS_HT - 4  ; clip against top
  bcc @not_above
  lda #PF_VIS_HT - 4
@not_above:
  sta 1   ; line currently being checked
  sta dirty_bottom,x  ; redraw lines that were checked

  lda #4
  sta 0   ; number of lines to check

  lda #0
  sta 2   ; number of lines found

  ; search each line
@lineloop:
  ldy 1
  lda rowOffsets,y
  tay
  ldx #10
@chkblkloop:
  lda (4),y
  beq @not_this_line
  iny
  dex
  bne @chkblkloop
  inc 2

  ldx #5  ;fill this row with CLEARED_ROW_TILE
  lda #CLEARED_ROW_TILE
@fillrowloop:
  dey
  sta (4),y
  dey
  sta (4),y
  dex
  bne @fillrowloop
@not_this_line:
  inc 1
  dec 0
  bne @lineloop
  ldx cur_turn
  lda 2
  bne @yes_lines
  sta cur_combo,x  ; if no lines, clear the combo
  jmp set_garbage_state

@yes_lines:
;;; do stuff with lines
  lda #STATE_POST_CLEAR
  sta cur_state,x
  lda #POST_CLEAR_DELAY
  sta state_timer,x

  lda 2
  pha
  cmp #4
  lda #4
  adc #0
  jsr start_sound
  pla
  sta 2
  ldx cur_turn

.if PLAY_4LINE_SOUND
  lda 2
  cmp #4
  bne @no_tetris_sound
  lda #DMC_4LINES
  jsr play_dmc
  ldx cur_turn
@no_tetris_sound:
.endif

  ; add lines to cumulative lines
  lda 2
  tay
  clc
  adc lines_lo,x
  sta lines_lo,x
  bcc @no_lines_carry
  inc lines_hi,x
@no_lines_carry:

  ; save whether there was already a combo
  ; so that we can display the proper top text
  lda cur_combo,x
  sta last_combo,x
  ; add lines to combo
  tya
  clc
  adc cur_combo,x
  sta cur_combo,x

  ; add lines * combo to score
  tya
@score_add_loop:
  clc
  lda cur_combo,x
  adc score_lo,x
  sta score_lo,x
  bcc @no_score_carry
  inc score_hi,x
@no_score_carry:
  dey
  bne @score_add_loop

  ; if garbage is enabled, add garbage to other player's side
  lda garbage_enabled
  beq @no_garbage
  txa        ; select other player
  eor #1
  tax
  ldy 2
  lda garbage_per_lines,y
  clc
  adc got_garbage,x
  sta got_garbage,x
  ldx cur_turn
@no_garbage:
  ; fall through to

write_combo_text:
  lda #(PF_VIS_HT - 1) * 10 + 1
  sta 4
  ldx cur_turn
  txa
  ora #4
  sta 5
  ldy cur_combo,x
  lda last_combo,x
  beq @write_pfbg_no_combo
    ldy #5
  @write_pfbg_no_combo:
  tya
  asl a
  asl a
  asl a
  tax
  dex
  ldy #7
  @copy_line_clear_name:
    lda (4),y
    bmi :+
      lda lineClearTypeText,x
      sta (4),y
    :
    dex
    dey
    bpl @copy_line_clear_name
  ldx cur_turn
  lda last_combo,x
  beq @no_write_pfbg_name

  ; BCD-convert the combo count to a temp buffer at $0100 first
  lda cur_combo,x
  sta dividend
  lda #0
  sta dividend+1
  sta $100  ; clear out the high digit
  sta 4
  lda #$01
  sta 5
  ldy #2
  jsr decimalize
  lda cur_turn
  ora #4
  sta 5
  lda #(PF_VIS_HT - 1) * 10 + 1
  sta 4
  ldy #1
  @copy_combo_count:
    lda (4),y
    bmi :+
      lda $100,y
      sta (4),y
    :
    dey
    bpl @copy_combo_count


@no_write_pfbg_name:
  rts


;;; post_clear
;   For each row that begins with a CLEARED_ROW_TILE,
;   set it to an EMPTY_ROW_TILE, which should look like a blank tile
;   but be of a different number.
post_clear:
  ldx cur_turn
  dec state_timer,x
  beq @ready
  rts
@ready:
  txa
  ora #>FIELD_1P
  sta 5
  lda #0
  sta 4

; setup next state
  lda #STATE_FALL
  sta cur_state,x
  lda #1
  sta state_timer,x
  lda piece_y,x
  bpl @not_below
  lda #0
  sta piece_y,x
@not_below:
  sta dirty_bottom,x

  lda #PF_HT - 1 ;current line
  sta 0
@lineloop:
  ldy 0
  lda rowOffsets,y
  tay
  lda (4),y
  cmp #CLEARED_ROW_TILE
  bne @not_this_row
  ldx #5
  lda #EMPTY_ROW_TILE
@blkloop:
  sta (4),y
  iny
  sta (4),y
  iny
  dex
  bne @blkloop
@not_this_row:
  dec 0
  bpl @lineloop

  rts


move_down:
  jsr erasePFBGText
  ldx cur_turn
  dec state_timer,x
  beq @ready
  rts
@ready:
  txa
  ora #>FIELD_1P
  sta 5         ; 4: dst field base
  sta 3         ; 2: src field base
  lda #0
  sta 4
@find_cleared_row:
  lda piece_y,x
  sta dirty_bottom,x
  cmp #PF_HT       ; have we made everything fall up to the top?
  bcc @still_got_more
  jmp set_garbage_state
@still_got_more:
  tay           ; is row y cleared?
  lda rowOffsets,y
  tay
  lda (4),y
  cmp #EMPTY_ROW_TILE
  beq @got_one
  inc piece_y,x
  bne @find_cleared_row
@got_one:       ; so we've found a cleared row.
                ; now move everything above it down one row.
  lda #10
  sta 2
@scroll_loop:
  lda (2),y
  sta (4),y
  iny
  lda (2),y
  sta (4),y
  iny
  cpy #(PF_HT - 1) * 10
  bcc @scroll_loop
  lda #0
@clear_top_loop:
  sta (4),y
  iny
  sta (4),y
  iny
  cpy #PF_HT * 10
  bcc @clear_top_loop

  ; only play the line clear landing sound
  ; if the row that was moved into this row
  ; also needs to be cleared
  lda piece_y,x
  tay           ; is row y cleared?
  lda rowOffsets,y
  tay
  lda (4),y
  cmp #EMPTY_ROW_TILE
  beq @no_land_sound
  lda #2
  jsr start_sound
  ldx cur_turn
@no_land_sound:

  lda #MOVE_DOWN_DELAY
  sta state_timer,x
  jmp write_combo_text

;;; move_up
;   Moves blocks in well up by 1 row in response to garbage.
move_up:
  ldx cur_turn
  dec state_timer,x
  beq @ready
  rts
@ready:
  txa
  ora #>FIELD_1P
  sta 5        ; 4: src field base
  sta 3        ; 2: dst field base
  lda #10
  sta 2
  lda #0
  sta 4
  sta dirty_bottom,x
  ldy #(PF_HT - 1)*10-1
@scroll_loop:
  lda (4),y
  sta (2),y
  dey
  cpy #$ff
  bne @scroll_loop

; add garbage line
  ldy #9
  lda #GARBAGE_TILE
@add_garbage_loop:
  sta (4),y
  dey
  bpl @add_garbage_loop

; find where to put a hole in the garbage
  ldy #3
  jsr random
  lda rand3
  and #$07
  tay
  iny

; knock out a hole in the garbage
  lda #0
  sta (4),y

  lda #6
  jsr start_sound
  ldx cur_turn
  dec got_garbage,x
  ; and fall into set_garbage_state

;;; set_garbage_state
;   Sets the state of a player to receiving garbage or new piece
;   depending on whether the player has garbage to receive.
;   in: x: player number
set_garbage_state:
  lda #MOVE_DOWN_DELAY
  sta state_timer,x
  lda got_garbage,x
  bne @is_garbage
  lda #STATE_NEW_PIECE
  sta cur_state,x
  lda rotation_system,x
  beq @center_are
  lda #POST_CLEAR_DELAY
  sta state_timer,x
  rts

@center_are:
  ; In CENTER, introduce some entry delay (ARE)
  ; around the transition from 1G to 5G
  lda piece_yspeed_whole,x
  asl a
  cmp #POST_CLEAR_DELAY
  bcc :+
    lda #POST_CLEAR_DELAY
  :
  sta state_timer,x
  rts

@is_garbage:
  lda #STATE_MOVE_UP
  sta cur_state,x
  rts



redraw_calc:

.if REDRAW_ALL_THE_TIME
  ldx #1
@reset_redraw:
  lda dirty_bottom,x
  cmp #22
  bne @skip_reset_redraw
  lda #0
  sta dirty_bottom,x
@skip_reset_redraw:
  dex
  bpl @reset_redraw
.endif

;;; Calculate number of lines that need to be redrawn
  ldx #1
@calc_hi:
  lda #PF_HT + 2

  ; don't copy the top two lines in attract mode
  ldy attractDataPtr+1
  beq @not_attract
  lda #PF_VIS_HT
@not_attract:
  
  sec
  sbc dirty_bottom,x
  bcs :+
  lda #0
:
  sta rows_left,x
  dex
  bpl @calc_hi
  
;;; Reduce this to fit redraw engine speed, if necessary
  clc
  lda rows_left
  adc rows_left+1
  sec
  sbc #MAX_ROWS_PER_FRAME*2
  bcc @no_reduce
  beq @no_reduce
;;; Now, A holds the total number of rows over the maximum.
;   Divide it by 2, rounding up.  Then reduce the rows to
;   draw on each side by that many.
  lsr a
  adc #0
  sta 0
  sec
  lda rows_left
  sbc 0
  bcs :+
  lda #0
:
  sta rows_left
  sec
  lda rows_left+1
  sbc 0
  bcs :+
  lda #0
:
  sta rows_left+1
@no_reduce:
  rts



;;; copy_pal:
;   Copy current palette from RAM to PPU.
;
copy_pal:
  lda #$3f
  sta PPUADDR
  ldx #0
  stx PPUADDR
  clc
  ldy palbuf  ; don't reload it every time to shave some cycles
@loop:
  sty PPUDATA
  lda palbuf+1,x
  sta PPUDATA
  lda palbuf+2,x
  sta PPUDATA
  lda palbuf+3,x
  sta PPUDATA
  sty PPUDATA
  lda palbuf+5,x
  sta PPUDATA
  lda palbuf+6,x
  sta PPUDATA
  lda palbuf+7,x
  sta PPUDATA
  txa
  adc #8
  tax
  cpx #32
  bcc @loop
  rts


;;; fbi_warning:
;   Loads a static map and waits for a player to press start.
;   0: address of packed data
;   Y: maximum number of seconds to wait
.global fbi_warning
fbi_warning:
  sty $08
  ldx #0
  stx PPUMASK
  lda #$20
  sta PPUADDR
  stx PPUADDR
  jsr PKB_unpackblk

  lda #0
  sta vblanked
  sta SPRADDR
  lda #>OAM
  sta SPRDMA
  lda #PPUCTRL_NMI
  sta PPUCTRL
  ; wait for vblanked to become 1, which can happen only once
  ; the NMI handler has run at the start of vblank
:
  lda vblanked
  beq :-

  jsr copy_pal
  lda #%00011110  ; show bg and sprites
  sta PPUMASK
  lda #0
  sta vblanked
  sta PPUSCROLL
  sta PPUSCROLL
  ldy $08

;;
; A: nametable address to wait on
; Y: maximum number of seconds to wait, or 0 for unlimited
startwait:
  sty $08
  ora #PPUCTRL_NMI
  sta PPUCTRL
  lda #FPS
  sta $09
  lda #$ff
  sta 0
  lda #0
  sta vblanked
@loop:
  lda vblanked
  beq @loop
  dec vblanked

  ldx #0
  stx SPRADDR  ; copy sprites EVERY FRAME
  lda #>OAM
  sta SPRDMA

  inc rand1
  bne :+
    inc rand2
  :
  jsr read_pads
  jsr update_sound

  lda $08
  beq @no_auto_end
    dec $09
    bne @no_auto_end
    lda #FPS
    sta $09
    dec $08
    beq @got

  @no_auto_end:
  lda new_keys
  ora new_keys+1
  and #%00010000
  beq @loop
@got:
  ; clock the accumulated randomness into the RNG
  ldy #8
  jmp random


;;
; Erases the text behind the top of the playfield.
; Does not erase EMPTY_ROW_TILEs in the far left column,
; which the line clear gravity code uses to mark lines
; that have been cleared.
; @param X player number
erasePFBGText:
  lda cur_turn
  ora #>FIELD_1P
  sta 5
  lda #(PF_VIS_HT - 2) * 10
  sta 4
  ldy #2*10-1
  @loop:
    lda (4),y
    bmi @isBlock
    cmp #EMPTY_ROW_TILE
    bne @isSpace
    cpy #10
    beq @isBlock
    cpy #0
    beq @isBlock
    @isSpace:
      lda #0
      sta (4),y
    @isBlock:
    dey
    bpl @loop
  rts



;;; div10
;   Given a 16-bit number in dividend, divides it by ten and
;   stores the result in dividend.
;   out: A: remainder; X: 0; Y: unchanged
div10:
  ldx #16
  lda #0
@divloop:
  asl dividend
  rol dividend+1
  rol a
  cmp #10
  bcc @no_sub
  sbc #10
  inc dividend
@no_sub:
  dex
  bne @divloop
  rts


;;; decimalize
;   Given a pointer to a text field, turns
;   in: 4: text field ptr
;       y: field length
;       dividend: number to decimalize
;   out: written
decimalize_loop:
  jsr div10
  cmp #0
  bne @nonzero
  ldx dividend
  bne @nonzero
  ldx dividend+1
  bne @nonzero
  rts
@nonzero:
  ora #'0'
  sta (4),y
decimalize:
  dey
  bpl decimalize_loop
  rts

;;; update_score
;   Updates the display of score and lines for the current player.
;   in: cur_turn: current turn
update_score:

  ; only update the score when not in attract mode
  ldx attractDataPtr+1
  beq @not_attract
  rts
@not_attract:
  ldx cur_turn
  lda dirty_bottom,x  ; if not dirty, mark them dirty
  cmp #PF_VIS_HT
  bcc @no_reset_dirty
  lda #PF_VIS_HT
  sta dirty_bottom,x
@no_reset_dirty:
  txa  ; find field pointer
  ora #>FIELD_1P
  sta 5

  lda #(PF_HT + 1) * 10 + 8  ; draw score ones place
  sta 4
  ldy #0
  sty dividend+1

  ; If the score has a hundreds part, zero out the ones. 
  lda score_lo,x
  ora score_hi,x
  beq @no_hundreds
  lda #'0'     ; zero out tens place if it doesn't
  sta (4),y
  iny
  sta (4),y
@no_hundreds:
  lda score_ones,x
  sta dividend
  ldy #2
  jsr decimalize

  ldx cur_turn  ; draw score hundreds place
  lda #(PF_HT + 1) * 10 + 3
  sta 4
  lda score_lo,x
  sta dividend
  lda score_hi,x
  sta dividend+1
  ldy #5
  jsr decimalize
  
  ldx cur_turn  ; draw lines place
  lda #PF_HT * 10 + 5
  sta 4
  .if SPEED_CURVE_REPLACES_LINES
    lda piece_yspeed,x
    sta dividend
    lda piece_yspeed_whole,x
    sta dividend+1
  .elseif LOCK_DELAY_REPLACES_LINES
    lda lock_delay,x
    sta dividend
    lda #0
    sta dividend+1
  .else
    lda lines_lo,x
    sta dividend
    lda lines_hi,x
    sta dividend+1
  .endif
  ldy #5
  jmp decimalize



loadWellAttribute:

  ; don't touch the well attribute if vblank time is
  ; going to be spent on blitting the playfields
  lda dirty_bottom
  cmp #PF_VIS_HT
  bcc @doNothing
  lda dirty_bottom+1
  cmp #PF_VIS_HT
  bcc @doNothing
  ldx #1
@xloop:
  lda well_attr_dirty,x
  bne loadWellAttributeNow
  dex
  bpl @xloop
@doNothing:
  rts

loadWellAttributeNow:
  stx 0
  lda #0
  sta well_attr_dirty,x
  lda cur_state,x
  cmp #STATE_GAMEOVER
  bne @notGameOver
    lda #0
  @notGameOver:
  cmp #STATE_START_GAME
  bcs @notPreGame
    lda #0
  @notPreGame:
  cmp #0
  beq @notGameOverOrInactive
    lda #2
  @notGameOverOrInactive:
  ora 0
  tay
  lda wellAttr0,y
  sta 0
  lda wellAttr12,y
  sta 1
  lda wellAttr3,y
  sta 2

  ; actually copy attributes
  lda #$23
  sta PPUADDR
  txa
  asl a
  asl a
  ora #$D0
  sta PPUADDR
  ldy #5
  :
    lda 0
    sta PPUDATA
    lda 1
    sta PPUDATA
    sta PPUDATA
    lda 2
    sta PPUDATA
    lda PPUDATA
    lda PPUDATA
    lda PPUDATA
    lda PPUDATA
    dey
    bne :-
  rts

.segment "RODATA"

.align 256
gameover_text:
  .byt "          "
  .byt " TO PLAY  "
  .byt "PRESS A+B "
  .byt "          "
  .byt "          "
  .byt "          "
  .byt "          "
  .byt "          "
  .byt "          "
  .byt "          "
  .byt "          "
  .byt "          "
  .byt "yv z z|tuz"  ; OVER
  .byt "uz uuzs }v"  ; OVER
  .byt "vyz zzsqwy"  ; OVER
  .byt "          "
  .byt "yvz uu u}|"  ; GAME
  .byt "u|z|uuuuwq"  ; GAME
  .byt "vyxst}xuws"  ; GAME
  .byt "          "
board_top:
  .byt "LINES    0"
  .byt "SC       0"
gamepal:
  .byt $0f,$00,$10,$30,$0f,$06,$16,$26,$0f,$02,$12,$22,$0f,$12,$27,$2A
  .byt $0f,$00,$10,$30,$0f,$00,$10,$30,$0f,$00,$10,$30,$0f,$00,$10,$30
board_x:
  .byt 2, 20
.align 256
menu_text:
  .byt "          "
  .byt "A:   START"
  .byt $1c,$1d,$1e,$1f,":  SET"
  .byt "          "
  .byt "          "
  .byt "          "
  .byt "          "
  .byt " MUSIC    "
  .byt "          "
  .byt "          "
  .byt " GARBAGE  "
  .byt "          "
  .byt "          "
  .byt " SPEED    "
  .byt "          "
  .byt "          "
  .byt " ROTATION "
  .byt "          "
  .byt " OPTIONS: "
  .byt "          "
;     -- vanishzone --
  .byt " TO LJ65! "
  .byt " WELCOME  "

pf_side_borders:
  .byt $80,$6b,$00,$00,$00,$00,$00,$00
  .byt $00,$00,$00,$00,$6c,$80,$80,$80
  .byt $80,$80,$80,$6b,$00,$00,$00,$00
  .byt $00,$00,$00,$00,$00,$00,$6c,$80

.align 256

ppubaselo:
  .byt $60,$40,$20,$00,$e0,$c0,$a0,$80
  .byt $60,$40,$20,$00,$e0,$c0,$a0,$80
  .byt $60,$40,$20,$00,$00,$00,$00,$80
  .byt $60
ppubasehi:
  .byt $23,$23,$23,$23,$22,$22,$22,$22
  .byt $22,$22,$22,$22,$21,$21,$21,$21
  .byt $21,$21,$21,$21,$00,$00,$00,$20
  .byt $20

rowOffsets:
  .byt   0, 10, 20, 30, 40, 50, 60, 70, 80, 90
  .byt 100,110,120,130,140,150,160,170,180,190
  .byt 200,210,220,230,240

garbage_per_lines:
  .byt 0, 0, 1, 2, 4
speed_names:
  .byt "SLOW MEDFAST 20G"
rotation_names:
  .byt "  CENTER", "  BOTTOM"
song_names:
  .byt "     OFF", "    TEMP", "   K.231"

initial_speeds:
  .byt     250/FPS, 1800/FPS, 12000/FPS, 0
initial_speeds_whole:
  .byt     0, 0, 0, 20

on_off_text:
  .byt " OFF ON!"

lineClearTypeText:
  .byt " SINGLE "
  .byt " DOUBLE "
  .byt " TRIPLE "
  .byt "HOME RUN"
  .byt "   COMBO"

wrong_region_message:
;        01234567890123456789012345678901
  .byte " REGION CODING FAILED!          "
  .byte " THIS GAME PAK IS DESIGNED FOR  "
.ifdef PAL
  .byte " PAL NES CONSOLES ONLY.         "
.else
  .byte " NTSC NES AND FAMICOM ONLY.     "
.endif
  .byte 0

wellAttr0:
  .byt $11, $AA, $DD, $AA
wellAttr12:
  .byt $00, $00, $FF, $FF
wellAttr3:
  .byt $55, $88, $55, $BB


.segment "VECTORS"

  .addr nmihandler, main, irqhandler

